/*
 *  First Anthrox GameBoy Advance demo
 *  Star sprites are 16 colors
 *  Logo is 16 colors
 *  Scroll Sprites are 256 colors and rotate
 *
 *  Code by -Pan-/Atx
 *
 *  Based on Warder1's source code for background and sprite
 *
*/

#define DISPCNT     *(u32*)0x04000000   // Define video display control
#define BG0CNT      *(u16*)0x04000008   // Define BG0 control
#define BG1CNT      *(u16*)0x0400000a   // Define BG1 control
#define BG0HOFS     *(u16*)0x04000010   // BG0 Horizontal offset
#define BG0VOFS     *(u16*)0x04000012   // BG0 Vertical offset
#define DMA3SAD     *(u32*)0x040000d4   // Define DMA 3 Source
#define DMA3DAD     *(u32*)0x040000d8   // Define DMA 3 Destination
#define DMA3CNT     *(u32*)0x040000dc   // Define DMA 3 data amount
#define OAM     ((Sprite*)0x07000000)   // Setup pointer to OAM ram
#define OBJCHARDATA ((u8*)0x06010000)   // Setup pointer to OBJ graphics data
#define OBJ_PLTT    ((u16*)0x05000200)  // Setup pointer to OBJ palette

//some type definitions
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned long u32;
typedef short s16;
//define a sprite as a structure
typedef struct _Sprite { u16 Attrib0, Attrib1, Attrib2, RotateScale; } Sprite;

// include the sinus data
#include "sinus.c"
// need this for RAND() function
#include <stdlib.h>
#include <math.h>

//Add graphics and pallete from data.asm
extern u8 logo1_gfx;                    // 16 color logo graphics
extern u16 logo1_pal;                   // 16 color palette
extern u8 logo1_map;                    // tile map
extern u16 sprite_pal;                  // 16 color sprite palette
extern u8 sprite_gfx;                   // 16 color sprite graphics


//Set up pointer to background palette RAM
u16 *pal=(u16*)0x5000000;
u16 *tmp_pal;
//Set up pointer to background gfx VRAM
u8 *tiles=(u8*)0x6004000;
u8 *bg_data;
//Set up pointer to background map VRAM
u16  *m0=(u16*)0x6000000;
u16 *bg_map;

// create some variables to use
u16 x,y;
u16 x1, y1, z1;
u32 x32,y32;

u8 H_sinus_offset;
u8 V_sinus_offset;

// Create Star Field Arrays
u16 star_z[80];
u16 star_x[80];
u16 star_y[80];
u8 scroll_text[] = "                              Hey cool fans!  -Pan- at the keys with this simple intro. More coming soon, like sprite rotation and more backgrounds and other neat-o things!     Wrap time is here!    \x0                              ";
u16 scroll_offset;
u16 scroll_offset2;
u16 scroll_x;
u16 scroll_sinus_offset;
u8 tempk;
u16 templ;
u16 starspeed;


void DMA3Call(u32 Src,u32 Dst,u32 Cnt)              //used to load our sprite info into vram
{
	DMA3SAD = Src;									//where the data is
	DMA3DAD = Dst;									//where the data goes
	DMA3CNT = Cnt;									//how much data should be loaded
}


void SetupScreen()
{
        u16 i;                     //define some temp variables

        bg_data = (u8*)(&logo1_gfx);                  //have bg_data point to our .raw defined in data.asm
        tmp_pal = (u16*)(&logo1_pal);                  //have tmp_pal point to our .pal defined in data.asm
        bg_map = (u16*)(&logo1_map);

        //for (i=0;i<256;i++) pal[i]=tmp_pal[i];  //load the pallette info into the desired address
        DMA3Call((u32)tmp_pal, (u32)pal, 0x84000010);  // DMA the palette


        //for(i=0;i<64*600;i++) tiles[i]=bg_data[i];
        //load the bg tiles into the desired address
        // DMA the BG Graphics
        DMA3Call((u32)bg_data, (u32)tiles,0x84000000+(4800)/2);  // DMA graphics

        //loop to load in the map data
        for (i=0;i<640;i++) m0[i]= bg_map[i];

        //DMA Sprite Colors
        // for (i=0;i<256;i++) OBJ_PLTT[i]=rand();
        DMA3Call((u32)&sprite_pal, (u32)OBJ_PLTT, 0x84000020);


        //DMA Sprite Graphics
        DMA3Call((u32)&sprite_gfx, (u32)OBJCHARDATA, 0x84000000+((6144+128)/2));
}

void SetupSprites()
{
        // set up star field x, y, z coords
        u8 k;
        u16 i;
        for ( i=0;i<80;i++) {
         k = rand();
         star_y[i] = (u16)(k/3.5)+4;
         k = rand();
         star_x[i] = (u16)(k/3.5)+4;
         k = rand();
         star_z[i] = (u16)k+2;
        }

        // push all sprites under the screen so they aren't visible
        for  (i=0;i<128;i++) OAM[i].Attrib0=168;
}



void Init()
{
    BG0CNT = 0x4004;       //turn on BG0CNT.  Also tells where the map and tile locations are for this bkg

    DISPCNT = 0x1140;      //sets the display mode to display BKG0
}

void VBlank()
{
    // Wait For Video Blank Interval to get in sync with screen
  __asm
   {
   /*   If you aren't using much CPU time you may find that it will
    * keep updating whenever its on line 160. enable this routine to correct timing
    * if you have that problem
    *
      mov     r0, #0x4000006
    scanline_wait1:
        ldrh   r1, [r0]
        cmp    r1, #160
      beq     scanline_wait1
    */

     mov     r0, #0x4000006
   scanline_wait2:
     ldrh	r1, [r0]
     cmp	r1, #160
    bne     scanline_wait2
   }
}


// This will take the star field data from the arrays
// and calculate the projection for one quadrant, figure out which star sprite
// to use (there are 4, from tiny dot to a small ball), and reset the x,y coords if the
// star goes off the screen
void calcstar(u16 t)
{
    y1 = star_y[t];
    x1 = star_x[t];
    z1 = star_z[t];

    // Get Original X position, multiply by 64, then divide by Z
    x32 = (u32)(((x1<<6) / z1) );
    // Get Originl Y position, multiply by 64, then divide by Z
    y32 = (u32)(((y1<<6) / z1) );

    x=(u16)x32;
    y=(u16)y32;

    // Screen is 240*160, if Sprite X > half of 240 then it's off the screen
    // and needs to be reset
    if ( x32>120)
    {
         x=248;
         z1=255;
         tempk = rand();
         star_y[t] = (u16)(tempk)+1;
         tempk = rand();
         star_x[t] = (u16)(tempk)+1;
         tempk = rand();
    }

    // Screen is 240*160, if Sprite Y > half of 160 then it's off the screen
    // and needs to be reset
    if ( y32>80)
    {
         y=88;
         z1=255;
         tempk = rand();
         star_y[t] = (u16)(tempk)+1;
         tempk = rand();
         star_x[t] = (u16)(tempk)+1;
         tempk = rand();
    }

    // move the Star's Z position
    z1 = z1 - starspeed;

    // the Z range goes from starspeed to 255  , 0 is not a good value!!
    if ( z1<starspeed)
    {
         z1=255;
         tempk = rand();
         star_y[t] = (u16)(tempk)+1;
         tempk = rand();
         star_x[t] = (u16)(tempk)+1;
         tempk = rand();
    }

    // Save the new Z Position
    star_z[t] = z1;

}

// Create 80 stars, 20 stars for each of the 4 quadrants
void DoStars()
{
    u16 t;
    u16 starframe;
    //quadrant 2 (neg x, pos y)
    for ( t=0;t<20;t++) {
    calcstar(t);
    x = 120-x;
    y += 80;
    starframe = z1>>6;   // z1>>6 = z1/64

    OAM[t].Attrib0 = y;
    OAM[t].Attrib1 = x;
    OAM[t].Attrib2 = 0x1400+starframe;
    }
    // Quadrant 1 (pos x, pos y)
    for ( ;t<40;t++) {
    calcstar(t);
    x = 120+x;
    y += 80;
    starframe = z1>>6;

    OAM[t].Attrib0 = y;
    OAM[t].Attrib1 = x;
    OAM[t].Attrib2 = 0x1400+starframe;
    }

    //quadrant 3  (neg x, neg y)
    for ( ;t<60;t++) {
    calcstar(t);
    x = 120+x;
    y = 80-y;
    starframe = z1>>6;

    OAM[t].Attrib0 = y;
    OAM[t].Attrib1 = x;
    OAM[t].Attrib2 = 0x1400+starframe;
    }

    //quadrant 4  (pos x, neg y)
    for ( ;t<80;t++) {
    calcstar(t);
    x = 120-x;
    y = 80-y;
    starframe = z1>>6;

    OAM[t].Attrib0 = y;
    OAM[t].Attrib1 = x;
    OAM[t].Attrib2 = 0x1400+starframe;
    }

}

void DoLogoSinus()
{
    //trapse through the sinus data
    H_sinus_offset +=1;
    V_sinus_offset +=1;

    // Add some extra randomness to the sinus
    // Since the X sinus is 256 bytes and the Y sinus is 160 bytes
    // they will be out of sync, by adding this line we can make even more
    // out of sync
    if ( H_sinus_offset==0) V_sinus_offset +=1;

    if ( V_sinus_offset > 159) V_sinus_offset = 0;
    // Write to Background 0 scroll register
    BG0HOFS = sinus_data[H_sinus_offset]-128;
    BG0VOFS = sinus_data_V[V_sinus_offset]-128;
}

void DoScroll()
{
        u16 i;
        u16 j;
        u16 k;
        // Sinus Scrolls are easy, but I did it the hard way. This is NOT a good example!
        // This is very dirty, changing the sinus offset (the k variable) to get a
        // steeper angle gets you lousy results.
        k=scroll_offset2;

        for ( i=0;i<32;i++) {
            j = scroll_sinus[(scroll_sinus_offset+k) & 255] | 0x2300;
            k = (k+4) & 255;
            OAM[81+i].Attrib0= j;
            OAM[81+i].Attrib1= i<<9 |((i<<3)-(scroll_x+8)+scroll_offset*8) & 511 ;
            OAM[81+i].Attrib2= 0x1000  | ((scroll_text[i+scroll_offset]-32)<<1)+4;
            //+64 = COS
            OAM[(i*4)].RotateScale = scroll_sin[(scroll_sinus_angle[(scroll_sinus_offset+k)+64 & 255])+64  & 255];
            // SIN
            OAM[(i*4)+1].RotateScale =scroll_sin[(scroll_sinus_angle[(scroll_sinus_offset+k)+64 & 255]) & 255];
            // negative SIN
            OAM[(i*4)+2].RotateScale =~scroll_sin[(scroll_sinus_angle[(scroll_sinus_offset+k)+64 & 255])  & 255];
            // +64 = COS
            OAM[(i*4)+3].RotateScale =scroll_sin[(scroll_sinus_angle[(scroll_sinus_offset+k)+64 & 255])+64  & 255];



        }

        scroll_sinus_offset = (scroll_sinus_offset+1) & 255;
        scroll_x++;

        if ( (scroll_x&7) == 0) {
            scroll_offset++;
            scroll_offset2-=4;
            scroll_sinus_offset += 8;
        }

        if ( scroll_text[scroll_offset]==0) {
             scroll_offset=0;
             scroll_offset2=0;
             scroll_x = 0;
             scroll_sinus_offset = rand() & 255;
        }
}

void C_Entry()
{
    //Set up registers for screen and sprites
	Init();										//call init to set up our screen display
    // Copy graphics and palette
    SetupScreen();                                  //display the background
    // Fill the star field array and clear sprite table
    SetupSprites();

    // Init sinus offsets
    H_sinus_offset = 0;
    V_sinus_offset = 0;
    // Init scroll offsets
    scroll_x = 0;
    scroll_offset = 0;
    scroll_offset2=0;
    scroll_sinus_offset = 0;
    //Init Stars speed
    starspeed = 4;
	while (1)									//loop
    {
    // Wait for VBlank Interval
    VBlank();
    // Make the logo move around
    DoLogoSinus();
    // Move the sprites and put them on screen
    DoStars();
    // Make the Sprite Scroller scroll
    DoScroll();
    }

 }

